
import java.lang.Math;
import java.awt.Color;
import jrobots.simulation.simulationObjects.JRobot2010;
import jrobots.utils.*;

public class Alfred extends JRobot2010 {
    
    private Angle scanAperture; //momentary scanAperture, needed in next frame 
    private Angle scanDirection; //momentary scanDirection, needed in next frame 
    private Scan latestSuccessfulScan; //last scanned target
    private int latestSuccessfulScanCount; // number of scans since last target
    private Vector[] latestTargetPositions;
    private double[] latestTargetPositionTimes;
    private Angle addAngle; 
    private Vector[] newTargetPositions;
    
    @Override
    protected void init()
    {
		setNameColor(new Color((float) 1.0, (float) 1.0, (float) 1.0));
		setBodyColor(new Color((float) 1.0, (float) 0.0, (float) 0.0));
		setTurretColor(new Color((float) 1.0, (float) 0.0, (float) 0.0));
        scanAperture = getMaxScanAperture();
        scanDirection = getOrientation();
        setScanAperture(scanAperture);
        setScanDirection(scanDirection);
        latestSuccessfulScan = null;
        latestSuccessfulScanCount = 0;
 
        addAngle = new Angle(30, "Degreees");
        latestTargetPositions = new Vector[4];
        for(int i = 0; i <= 3; i++)
        {
        	latestTargetPositions[i] = null; 
        }
        latestTargetPositionTimes  = new double[4];
        newTargetPositions = new Vector[4];
        for(int i = 0; i <= 3; i++)
        {
        	newTargetPositions[i] = null; 
        }
    }
    
    @Override
    protected void actions()
    {
    	if(latestSuccessfulScan == null)
    	{
    		System.out.println("Alfred: Stop trying to hit me and hit me!");
    	}
        calibrate();
        
        collect();
        
        scan();
        
        shoot();
        
        drive();
    }
        
    /** 
     * Vorbereitungen treffen 
     */
    private void calibrate()
    {
    	if(latestSuccessfulScan != null)
    	{
    		setDropMineCommand(true);
    	}
    }    
    
    /**
     * Sensoren auswerten
     */
    private void collect()
    {
    	// Scanner nur auswerten, wenn gerade gescannt wurde
        if (isScanFromNow())
        {
            if (this.getLastScan().isTargetLocated())
            {
        		latestSuccessfulScan = this.getLastScan();
                for(int i = 3; i > 0; i--)
                {
                	latestTargetPositions[i] = latestTargetPositions[i - 1];
                	latestTargetPositionTimes[i] = latestTargetPositionTimes[i - 1];
                }
                
                latestTargetPositions[0] = new Vector(latestSuccessfulScan.scanDirection.add(scanAperture.mult(0.5)),
        										latestSuccessfulScan.distanceToTarget).add(getPosition());
                latestTargetPositionTimes[0] = latestSuccessfulScan.timeOfScan;
        		if(latestSuccessfulScanCount > 0)
        		{
        			latestSuccessfulScanCount = -1; 
                } else {
                	latestSuccessfulScanCount--;
                }
            } else {
            	if(latestSuccessfulScanCount < 0)
            	{
            		latestSuccessfulScanCount = 1;
            	} else {
            		latestSuccessfulScanCount++;
            	}
            }
        }
    }
    
    /**
     * Wohin soll der Scanner ausgerichtet werden?
     */
    private void scan()
    {    	
        if (isScanFromNow())
        {     	
        	if(latestSuccessfulScanCount == 0)
        	{
        		addAngle = new Angle(60, "Degrees");
        	} else {
        		int angle = Math.min( 30, 5 * ((int)(Math.abs(latestSuccessfulScanCount / 2.0) + 1)) );
        		addAngle = new Angle(angle ,"Degrees");
        	}
        	
        	if(latestSuccessfulScan == null)
        	{
	        	scanDirection = scanDirection.add(addAngle);
	        } else if(latestSuccessfulScanCount < 0) {
      	       	scanDirection = scanDirection.sub(addAngle);
	        } else {
	        	scanDirection = scanDirection.add(addAngle);	        	
	        }
	        setScanDirection(scanDirection);        
        } else {
            setScanAperture(scanAperture);
            setScanDirection(scanDirection);
        }
    }
    
    /**
     * Wohin soll geschossen werden?
     */
    private void shoot()
    {
    	if(latestSuccessfulScan != null)
    	{
    		if (Math.random()<= 0.8)
    		{
    			setLaunchProjectileCommand(latestSuccessfulScan.scanDirection.add(scanAperture.mult(Math.random()/20+0.53)));	
    		} else {
    			setLaunchProjectileCommand(latestSuccessfulScan.scanDirection.add(scanAperture.mult(Math.random()/40+0.33)));
    		}
    	}
    	setDropMineCommand(true);	
    }
    
    /**
     * Wohin faehrt JRobot?
     */
    private void drive()
    {
    	if(latestSuccessfulScan != null)
    	{
    		setAutopilot(latestSuccessfulScan.scanDirection, 1.0);

	    } 
    }
}



