// (c) by Stephan Lunowa, July 2012

import java.lang.Math;
import java.awt.Color;

import jrobots.simulation.simulationObjects.JRobot2010;
import jrobots.utils.*;

public class CryBot extends JRobot2010 {
    private static final long serialVersionUID = 1L;
    
    private static final double CRITICAL_DISTANCE = 15.0;
    private static final double SMALL_DISTANCE = 40.0;
    private static final double LONG_DISTANCE = 60.0;
    private static final double MAX_DISTANCE = JRobot2010.getMaxArenaDiameter();
    private static final double SHOT_TIME_DELAY = 0.2;
    private static final Angle MAX_SCAN_APERTURE = getMaxScanAperture();
    
    private Angle scanDirection; //momentary scanDirection, needed in next frame 
    private Scan[] lastTargetScans; // last Scans, when target was found
    private int lastSuccessfulScanCount; // Anzahl der Scans seit letzten Fund
    private Angle delayAngle; // Verschiebungswinkel durch Eigenbewegung
    private Angle addAngle; // Drehwinkel des Scanners 
    private double speed;
    private Angle driveDirection;
    private int randomAddDirection;
    private Vector[] newTargetPositions;
    private Vector[] minePositions;
    private Vector[] lastPositions;
    private int stopCount;
    private Angle[] targetDirections;
    private int targetLeftRotationsCount; // gegen Uhrzeigersinn
    
    @Override
    protected void init() {
        setBodyColor(new Color(0xb2, 0x78, 0x08));
        setNameColor(Color.RED);
        setTurretColor(new Color(0xb2, 0x78, 0x08));
        scanDirection = getOrientation();
        setScanAperture(MAX_SCAN_APERTURE);
        setScanDirection(scanDirection);
        delayAngle = new Angle(2.5, "Degrees");
        lastSuccessfulScanCount = 0;
        addAngle = new Angle(30, "Degreees");
        speed = 0.7;
        lastTargetScans = new Scan[3];
        newTargetPositions = new Vector[5];
        minePositions = new Vector[4];
        lastPositions = new Vector[3];
        stopCount = 0;
        randomAddDirection = 0;
        targetDirections = new Angle[2];
        targetLeftRotationsCount = 0;
    }
    
    @Override
    protected void actions() {
        calibrate();
        
        collect();
        
        scan();
        
        shoot();
        
        drive();
    }
        
    /** 
     * Vorbereitungen treffen 
     */
    private void calibrate() {

    	// Gegnerposition ermitteln
    	double x0, dx,
    		   y0, dy,
    		   t0, dt,
    		   t, vp;
    	
    	if(lastTargetScans[0] != null && lastTargetScans[1] != null) {
        	for(int i = 4; i > 0; i--) {
    			newTargetPositions[1] = newTargetPositions[0];
        	}
        	
        	vp = (lastTargetScans[0].distanceToTarget < LONG_DISTANCE)? 
        				getProjectileSpeed() : getBulletSpeed();

        	addDebugLine(lastTargetScans[0].estimatedTargetPosition(), lastTargetScans[1].estimatedTargetPosition());
    		if(lastTargetScans[2] == null) { // linear
    			x0 = lastTargetScans[1].estimatedTargetPosition().getX();
    			dx = lastTargetScans[0].estimatedTargetPosition().getX() - lastTargetScans[1].estimatedTargetPosition().getX();
    			y0 = lastTargetScans[1].estimatedTargetPosition().getY();
    			dy = lastTargetScans[0].estimatedTargetPosition().getY() - lastTargetScans[1].estimatedTargetPosition().getY();
    			t0 = lastTargetScans[1].timeOfScan;
    			dt = lastTargetScans[0].timeOfScan - lastTargetScans[1].timeOfScan;
    			
    			t = getTime();
    			double result, lastResult = Double.MAX_VALUE;
    			
    			result = t - (getTime()+ SHOT_TIME_DELAY) - Math.sqrt( 
    								(getPosition().getX() - x0 - dx/dt * (t - t0)) 
    								* (getPosition().getX() - x0 - dx/dt * (t - t0)) 
    								+ (getPosition().getY() - y0 - dy/dt * (t - t0)) 
    								* (getPosition().getY() - y0 - dy/dt * (t - t0))) / vp;
		
    			while( Math.abs(result) > 0.5 && Math.abs(lastResult) > Math.abs(result) ) {
    				t += 0.05;
    				lastResult = result;
    				result = t*t * ( (getPosition().getX() - x0 - dx/dt * (t - t0)) * (getPosition().getX() - x0 - dx/dt * (t - t0)) 
    							      + (getPosition().getY() - y0 - dy/dt * (t - t0)) * (getPosition().getY() - y0 - dy/dt * (t - t0)))
    						 - vp;
    			}
				newTargetPositions[0] = new Vector(x0 + dx/dt * (t - t0), y0 + dy/dt * (t - t0));
				
    		} else { // TO-DO: bilinear
    			x0 = lastTargetScans[1].estimatedTargetPosition().getX();
    			dx = lastTargetScans[0].estimatedTargetPosition().getX() - lastTargetScans[1].estimatedTargetPosition().getX();
    			y0 = lastTargetScans[1].estimatedTargetPosition().getY();
    			dy = lastTargetScans[0].estimatedTargetPosition().getY() - lastTargetScans[1].estimatedTargetPosition().getY();
    			t0 = lastTargetScans[1].timeOfScan;
    			dt = lastTargetScans[0].timeOfScan - lastTargetScans[1].timeOfScan;
    			
    			t = getTime();
    			double result, lastResult = Double.MAX_VALUE;
    			
    			result = t - getTime() - Math.sqrt( (getPosition().getX() - x0 - dx/dt * (t - t0)) 
													 * (getPosition().getX() - x0 - dx/dt * (t - t0)) 
													+ (getPosition().getY() - y0 - dy/dt * (t - t0)) 
													 * (getPosition().getY() - y0 - dy/dt * (t - t0))) / vp;
	
    			
    			while( Math.abs(result) > 0.5 && Math.abs(lastResult) > Math.abs(result) ) {
    				t += 0.05;
    				lastResult = result;
    				result = t*t * ( (getPosition().getX() - x0 - dx/dt * (t - t0)) * (getPosition().getX() - x0 - dx/dt * (t - t0)) 
    							      + (getPosition().getY() - y0 - dy/dt * (t - t0)) * (getPosition().getY() - y0 - dy/dt * (t - t0)))
    						 - vp;
    			}
    			newTargetPositions[0] = new Vector(x0 + dx/dt * (t - t0), y0 + dy/dt * (t - t0));  			
    		}
    	}
    	
    	if(newTargetPositions[4] != null)
    		setDebugLines(newTargetPositions);
    	
    	for(int i = 0; i < 2; i++) {
	    	if(newTargetPositions[i] != null) {
//	    		addDebugArrow(lastTargetScans[0].estimatedTargetPosition(), newTargetPositions[i]);
	    		addDebugCrosshair(newTargetPositions[i]);
	    	}
	    }
	    
//	    for(int i = 0; i < 4 && minePositions[i] != null ; i++) {
//	    	 addDebugCrosshair(minePositions[i]);
//	    }
//	    addDebugLine(getPosition(), getPosition().add(new Vector(10, 0)));
    	
    	if(targetDirections[0] != null && targetDirections[1] != null) {
    			addDebugArrow(lastTargetScans[0].estimatedTargetPosition(), 
						lastTargetScans[0].estimatedTargetPosition()
							.add(new Vector(10, 
									(targetDirections[0].add(targetDirections[1])
											.mult(0.5))).mult(1/1.5)));
    	}
    }    
    
    /**
     * Sensoren auswerten
     */
    private void collect() {
    	// aktuelle Pos speichern
    	for(int i = 2; i > 0; i--) {
    		lastPositions[i] = lastPositions[i - 1];
    	}
    	lastPositions[0] = getPosition();
    	
    	// Scanner nur auswerten, wenn gerade gescannt wurde
        if (isScanFromNow()) {
            if (this.getLastScan().isTargetLocated()) {
                for(int i = 2; i > 0; i--){
                	lastTargetScans[i] = lastTargetScans[i - 1];
                }
                lastTargetScans[0] = new Scan(getLastScan().scanDirection
                							  .add(getLastScan().scanAperture),
                							  getLastScan().scanAperture, 
                							  getLastScan().distanceToTarget, 
                							  getLastScan().scannerPosition,
                							  getLastScan().timeOfScan);
                if(lastTargetScans[1] != null) {
                	targetDirections[1] = targetDirections[0];
                	targetDirections[0] = lastTargetScans[0].estimatedTargetPosition()
                							.sub(lastTargetScans[1].estimatedTargetPosition())
                						  .getAngle();
                }
                
         		if(lastSuccessfulScanCount > 0) {
                	lastSuccessfulScanCount = -1; 
                } else {
                	lastSuccessfulScanCount--;
                }
            } else {
            	if(lastSuccessfulScanCount < 0) {
            		lastSuccessfulScanCount = 1;
            	} else {
            		lastSuccessfulScanCount++;
            	}
            }
        }
    }
    
    /**
     * Wohin soll der Scanner ausgerichtet werden?
     */
    private void scan() {    	
    	// Scanner neu ausrichten, wenn gerade gescannt wurde
        if (isScanFromNow()) {     	
        	if(lastTargetScans[0] == null) {
        		addAngle = new Angle(30, "Degrees");
        	} else {
        		int angle = Math.min( 30, 
        				(int)((4 - 2 * lastTargetScans[0].distanceToTarget / MAX_DISTANCE) 
        					  * (Math.abs(lastSuccessfulScanCount / 2.0) + 1)) );
        		addAngle = new Angle(angle ,"Degrees");
        	}
        	
        	
        	if(lastTargetScans[0] == null) {
	        	scanDirection = scanDirection.add(addAngle);
	        } else if(lastSuccessfulScanCount < 0) {
      	       	scanDirection = scanDirection.sub(addAngle);
	        } else {
	        	scanDirection = scanDirection.add(addAngle);	        	
	        }

	        setScanDirection(scanDirection);
	        
        } else {
            // Wenn noch nicht gescannt wurde, zur Sicherheit nochmal Befehl setzen
            setScanAperture(MAX_SCAN_APERTURE);
            setScanDirection(scanDirection);
        }
    }
    
    /**
     * Wohin soll geschossen werden?
     */
    private void shoot() {
    	
    	// Minen setzen
    	if(getTime() >= getTimeOfDropperReloaded() && speed > 0.0){
 			setDropMineCommand(true);
 			for(int i = 3; i > 0; i--) {
 				minePositions[i] = minePositions[i - 1];
 			}
 			minePositions[0] = getPosition();
 		} 

		// schieen
    	if(newTargetPositions[0] != null && getTime() >= getTimeOfLauncherReloaded()) {
    		if(lastTargetScans[0].distanceToTarget < LONG_DISTANCE) {
    			setLaunchProjectileCommand(newTargetPositions[0].sub(getPosition())
    									.getAngle() );//.add(new Angle(1 - 2*Math.random(), "Degrees")));
    		} else {
    			setLaunchBulletCommand(newTargetPositions[0].sub(getPosition())
    									.getAngle() );//.add(new Angle(0 - Math.random(), "Degrees")));
    		}
     	} else {
   			setLaunchBulletCommand(new Angle( Math.random()*360 , "Degrees"));
    	}
    }
    
    
    /**
     * Wohin faehrt der JRobot?
     */
    private void drive() {
    	if(HasStopped()) { // Panzersperre getroffen
    		driveDirection = getOrientation().add(new Angle(5, "Degrees"));
    		speed = -1.0;
    	} else if(newTargetPositions[0] != null && 
    				getPosition().distanceTo(newTargetPositions[0]) < CRITICAL_DISTANCE) {
    		driveDirection =  (lastTargetScans[0] == null)? getOrientation() : 
        							lastTargetScans[0].scanDirection.add(delayAngle)
        							.add(new Angle(60, "Degrees"));
    		speed = 1.0;  
	        
    		addRandomDriveAngle();    		
    		
    	} else {
    		speed += 0.1 * (Math.random() - 0.5);
	    	if(speed < 0.5) {
	    		speed = 0.5;
	    	} else if(speed < 0.8 && newTargetPositions[0] != null && 
    				getPosition().distanceTo(newTargetPositions[0]) > LONG_DISTANCE) {
	    		speed = 0.8;
	    	} else if(newTargetPositions[0] != null && 
    				getPosition().distanceTo(newTargetPositions[0]) > MAX_DISTANCE - 15) {
	    		speed = 1.0;
	    	} else if(speed > 1.0) {
	    		speed = 1.0;
	    	}
	    	
//	    	if(newTargetPositions[0] !=null) {
//	    		driveDirection = newTargetPositions[0].sub(getPosition()).getAngle();
//	    	} else {
	        	driveDirection =  (lastTargetScans[0] == null)? getOrientation() : 
	        							lastTargetScans[0].scanDirection.add(delayAngle);
//	    	}
	    	
	        addRandomDriveAngle();
        	attendMines();

    	}
    
    	setAutopilot(driveDirection, speed);
    }
    
    private void addRandomDriveAngle() {
    	if(randomAddDirection < 0) {
    		driveDirection.sub(new Angle(20 - 10 * Math.random(), "Degrees"));
    		randomAddDirection++;
    	} else if(randomAddDirection > 0) {
    		driveDirection.add(new Angle(20 - 10 * Math.random(), "Degrees"));
    		randomAddDirection--;
    	} else {
    		randomAddDirection = (Math.random() > 0.5)? 10 : -10;
    		addRandomDriveAngle();
    	}
    }
    
    private void attendMines() {
//    	double distance = 10000.0;
        for(int i = 0; i < 4; i++) {
        	if(minePositions[i] != null && getPosition().distanceTo(minePositions[i]) < SMALL_DISTANCE) {
//	        	if(getPosition().distanceTo(minePositions[i]) < distance && 
//	        	   Math.abs( minePositions[i].sub(getPosition())
//	        			   					 .getAngle().sub(driveDirection)
//	        			   					 .getValueAsDegrees()) < 90 ) {
//	        		distance = getPosition().distanceTo(minePositions[i]);
        			if(minePositions[i].sub(getPosition()).getAngle()
        					           .sub(driveDirection).getValueAsDegrees() < 30 &&
        			   minePositions[i].sub(getPosition()).getAngle()
        					           .sub(driveDirection).getValueAsDegrees() >= 0) {
        				driveDirection = driveDirection.add(new Angle(-40, "Degrees"));
        			} else if(minePositions[i].sub(getPosition()).getAngle()
					           				  .sub(driveDirection).getValueAsDegrees() > -30 &&
		        			  minePositions[i].sub(getPosition()).getAngle()
		        				              .sub(driveDirection).getValueAsDegrees() <= 0) {
           				driveDirection = driveDirection.add(new Angle(40, "Degrees")); 				
        			}
//        		}
    		}
    	}
    }
    
    private boolean HasStopped() {
    	if((lastPositions[0] != null && lastPositions[1] != null 
    			&& (lastPositions[0].getX() == lastPositions[1].getX() 
    					&& lastPositions[0].getY() == lastPositions[1].getY()) ||
    	   (lastPositions[1] != null && lastPositions[2] != null 
    			&& (lastPositions[1].getX() == lastPositions[2].getX() 
    					&& lastPositions[1].getY() == lastPositions[2].getY())))) {
    		stopCount = 20;
    		return true;
    	} else if(stopCount > 0) {
    		stopCount--;
    		return true;
    	} else {
    		return false;
    	}
   	
    }

}
