import java.lang.Math;
import java.util.LinkedList;
import java.awt.Color;
import jrobots.utils.*;

import jrobots.simulation.simulationObjects.JRobot2010;

/**
 * Sozusagen ein Starter-Toolkit.
 * <p>
 * Er enth&auml;lt schon einen rundum-Scanner und kann auch schon fahren.
 * <p>
 * ...Was? Er schie&szlig;t nicht?
 * <p>
 * Vielleicht sollte man das mal &auml;ndern. Und nicht vergessen: Gebt ihm auch
 * einen sch&ouml;nen Namen.
 */
public class Quitschie extends JRobot2010 {

	private Angle scanAperture; // momentary scanAperture, needed in next frame
	private Angle scanDirection; // momentary scanDirection, needed in next
									// frame

	final double MAX_TIME_LOST = 2;

	private int segment;
	private int segments;
	private int s = 5;
	private int d = 0;
	private double distanceToTarget;
	private int time; 
	private int i;

	private LinkedList<Scan> scans = new LinkedList<Scan>();

	/**
	 * Sensoren auswerten
	 */
	private void collect() {
		// Scanner nur auswerten, wenn gerade gescannt wurde
		if (isScanFromNow()) {
			if (getLastScan().isTargetLocated()) {
				scans.add(this.getLastScan());
				distanceToTarget = getLastScan().distanceToTarget;
				if (scans.size() > 2) {
					scans.removeFirst();
				}
			} else {
				double lastTime = (scans.size() != 0) ? scans.getLast().timeOfScan
						: 0;

				if (this.getTime() - lastTime > MAX_TIME_LOST) {
					scans.clear();
				}
			}
		}
	}

	/**
	 * Wohin soll der Scanner ausgerichtet werden?
	 */
	private void scan() {
		boolean foundEnemy = scans.size() != 0;
		if (isScanFromNow()) {
			if (!foundEnemy) {
				segment++;
				segment = segment % segments;

				scanAperture = getMaxScanAperture();
				scanDirection = (new Angle(360, "Degrees"))
						.mult((double) segment / segments);
			} else if (foundEnemy) {
				boolean lastScanFound = getLastScan().isTargetLocated();
				if (lastScanFound) {
					scanDirection = scanDirection.sub(new Angle(s, "d"));

				}

				else {
					scanDirection = scanDirection.add(new Angle(s, "d"));
				}
			}
		}

		setScanAperture(scanAperture);
		setScanDirection(scanDirection);
	}

	/**
	 * Wohin soll geschossen werden?
	 */
	private void shoot() {

		
		if (scans.size() == 0)
			return;

		Vector estTarget;
		Scan s = scans.getLast();

		if (distanceToTarget < this.getMaxArenaDiameter() / 2) {
			if (scans.size() == 2) {
				estTarget = LinearPredictor.predict(scans.get(0), scans.get(1),
						getTime() + 1);
				Vector relTarget = estTarget.sub(getPosition());

				Angle estAngle = relTarget.getAngle().add(
						new Angle(s.scanAperture.getValueAsDegrees(), "d"));
				this.setLaunchProjectileCommand(estAngle);

			} else {
				this.setLaunchBulletCommand(scanDirection);
			}
		} else {
			this.setDropMineCommand(true);
			
			this.setLaunchProjectileCommand(scanAperture.add(new Angle (d, "d")));
			
	// Feuerwaffe abfeuern
		}
	}

	/**
	 * Wohin f&auml;hrt der JRobot?
	 */
	private void drive() {
		if (scans.size() == 0)
			return;

		// in Richtung des letzten gescannten Gegners fahren
		setAutopilot(scans.getLast().scanDirection, 1.0);
	}

	@Override
	protected void actions() {
		calibrate(); // should be first

		collect(); // must be before scan

		scan(); // must be before drive

		shoot();

		drive();
	}

	@Override
	protected void init() {
		setBodyColor(new Color(0, 0, 100));
		setNameColor(new Color((float) 1.0, (float) 1.0, (float) 0.0));
		
		
		
		setAutopilot(getOrientation(), 1.0);
		scanAperture = getMaxScanAperture();
		scanDirection = getOrientation();
		setScanAperture(scanAperture);
		setScanDirection(scanDirection);
	}

	/**
	 * Hier berechnet der Bot aus den "Konstanten" der Umwelt abgeleitete Werte,
	 * die er in seinen Routinen <code>scan()</code>, <code>shoot()</code> und
	 * <code>drive()</code> verwendet.
	 */
	private void calibrate() {
		segments = (int) Math.ceil(360.0 / getMaxScanAperture()
				.getValueAsDegrees());
	}
}

